#define VGAPAINT_LOSSYLZ_C

#include <math.h>
#include <string.h>
#include <ctype.h>
#include <malloc.h>
#include <stdlib.h>

#include <a/types.h>

#include "video.h"
#include "pcx.h"
#include "io.h"
#include "asmfunc.h"
#include "toolicon.h"

ulong	HuffmanMakeEncodingTable(uchar *mem,ulong memsiz,ushort *encode_table,uchar *bitlen_table,int max_code);
void	HuffmanWriteNibbleTable(FILE *fo,uchar *data,ulong len);
void	HuffmanWriteBlock(FILE *fo,uchar *data,ulong len,ushort *encode_table,uchar *bit_table);
boolean	HuffmanReadNibbleTable(FILE *fi,uchar *table,ulong len);
void	HuffmanReconstructTable(uchar *table,uchar *lengths,ulong maxc);
boolean	HuffmanReadBlock(FILE *fi,uchar *mem,ulong len,uchar *decode_table,uchar *bitlen);

// Lossy-LZ compression algorithm:
// 1) Compress the picture using a lossy LZ algorithm.  Produce four blocks:
//    a block of literals, a block of distance codes, a block of length
//    codes, and a block of literal/replicate bits.
// 2) Huffman compress the literals.
// 3) Huffman compress the lengths.
// 4) Huffman pack the bit code distance low bytes.
// 5) Huffman pack the bit code distance high bytes.
// 6) Lossless LZ compress the literal/replicate bits.
//
// Note: When we huffman pack, we get a 256 nibble table of bit lengths.
//	 That's 128 bytes, which would end up 640 bytes total for all 5
//	 tables -- too much!  So we huffman pack the bit table too, which
//	 should always save a good amount of space (maybe as much as a
//	 sector).

struct LossyLZHeader {
	ulong	magic;
	ulong	header_size;
	ulong	width,height;
	ulong	literals;	// literal size
	ulong	runs;		// run size
	ulong	literal_pos;
	ulong	length_pos;
	ulong	bits_pos;
	ulong	distlo_pos;
	ulong	disthi_pos;
};

#define LLZH_MAGIC	'ZLPV'

//	int result = LossyLZCompress(uchar *block,ulong size,int threshold);
//
//	Attempt to LossyLZ compress a block.
//
//	Return codes:
//		positive:	new size of block
//		0:		compression successful, but no space gained
//		-1:		not enough memory

#define QUIT(x) { err=(x); goto cleanup0; }
#define MAX_LEN	(12)

int	LossyLZCompress(FILE *fo,uchar *block,ulong size,int threshold) {
	uchar	*literal_mem=NULL; 	// literal block memory
	uchar	*length_mem=NULL;  	// length block memory
	uchar	*dist_lo_mem=NULL;	// distance block memory - low byte
	uchar	*dist_hi_mem=NULL;	// distance block memory - high byte
	uchar	*bits_mem=NULL;		// bit memory
	ushort	*hencode_tbl=NULL;	// huffman encoding table - 256*5+16*5 entries
	uchar	*hbit_tbl=NULL;		// huffman bitlength table - 256*5+16*5 entries
	ulong	*square_tbl=NULL;

	uchar	*literals = literal_mem;
	uchar	*lengths = length_mem;
	uchar	*distlo = dist_lo_mem;
	uchar	*disthi = dist_hi_mem;
	uchar	*bits = bits_mem;
	ulong	*squarearray;

	int	err=0;
	int i;

	// allocate memory for blocks

	// Given a block of <x> bytes to compress:
	//	1) Literal block max size: x bytes
	//	2) Length block max size: (x+2)/3 bytes
	//	3) Distance block max size: 2(x+2)/3 bytes
	//	4) Literal/replicate block max size: (x+23)/24

	if (!(literals = literal_mem	= malloc(size				))) QUIT(-1);
	if (!(lengths  = length_mem	= malloc((size+2)/3			))) QUIT(-1);
	if (!(distlo   = dist_lo_mem	= malloc((size+2)/3			))) QUIT(-1);
	if (!(disthi   = dist_hi_mem	= malloc((size+2)/3			))) QUIT(-1);
	if (!(bits     = bits_mem	= malloc((size+23)/24			))) QUIT(-1);
	if (!(square_tbl = malloc(511*sizeof(ulong)))) QUIT(-1);
	if (!(hencode_tbl	= malloc((256+16)*5*sizeof(ushort)	))) QUIT(-1);
	if (!(hbit_tbl		= malloc((256+16)*5*sizeof(uchar)	))) QUIT(-1);

	squarearray = square_tbl+255;

	for(i=-255;i<=255;i++) squarearray[i]=i*i;

	threshold *= threshold;

	// PASS 1: LOSSY LZ COMPRESSION

	#define PRINT_INTERVAL	(1024)
	#define MAX_MATCH	(255+3)

	{
		uchar	*in_ptr = block;
		uchar	*in_limit = block + size;
		uchar	*pr_limit = block+PRINT_INTERVAL;
		uchar	*window;
		int	max_match;
		int	longest_match_len;
		uchar	*longest_match;
		long	bits_accum = 0x00800000L;

		// this is the lousiest LZ compression code ever, but oh well

		*literals++ = *in_ptr++;	// the first byte must be a copy

		do {
			// find the length of the block to search

			window		  = max(block,in_ptr - 4096);
			max_match	  = min(min(MAX_MATCH,in_limit - in_ptr),in_ptr+(MAX_MATCH-1)-block);
			longest_match_len = 0;
			do {
				// look for a match

				int therr=0;
				int err=0;
				int ml;
				int l = max_match;
				uchar *a = window;
				uchar *b = in_ptr;

				do {
					err = squarearray[(long)*a++ - (long)*b++];
					if (err>threshold) break;
					therr += err;
				} while(--l);

				// suitable match, construct "length"

				ml = ((max_match-l)<<22)+therr;
				if (ml>=longest_match_len) {
					longest_match_len = ml;
					longest_match = window;
					if (ml==(max_match<<22)) break;
				}
			} while(++window<in_ptr);

			bits_accum<<=1;
			longest_match_len>>=22;

			if (longest_match_len>2) {
				ulong dist = ((longest_match+4096)-in_ptr);

				*distlo++ = (uchar)dist;
				*disthi++ = (uchar)(dist>>8);

				*lengths++ = longest_match_len-3;
				memcpy(in_ptr,longest_match,longest_match_len);	// better lossy matches later
				in_ptr += longest_match_len;
				++bits_accum;
			} else {
				*literals++ = *in_ptr++;
			}
			if (bits_accum<0) {
				*bits++ = (uchar)bits_accum;
				bits_accum = 0x00800000L;
			}

			if (in_ptr>=pr_limit) {
				kbhit();
				SetTitle("Data:%ld Lit:%ld Runs:%ld",pr_limit-block,literals-literal_mem,lengths-length_mem);
				pr_limit+=PRINT_INTERVAL;
			}
		} while(in_ptr<in_limit);

		if (bits_accum>0x00800000L) {
			while(bits_accum>=0) bits_accum<<=1;
			*bits++ = bits_accum;
		}
	}

	// build a huffman table for literals

	{
		ulong	len,tlen=0;
		struct	LossyLZHeader llzh;

		if (-1==(len=HuffmanMakeEncodingTable(literal_mem,literals-literal_mem,hencode_tbl+   0,hbit_tbl+   0,256))) QUIT(-1);
		tlen+=len;

		Message("literal_mem: %d %d %d %d %d",
			literal_mem[0],
			literal_mem[1],
			literal_mem[2],
			literal_mem[3],
			literal_mem[4]);

		Message("literals: %ld->%ld",literals-literal_mem,len);

		if (-1==(len=HuffmanMakeEncodingTable(hbit_tbl   ,256                 ,hencode_tbl+1280,hbit_tbl+1280, 16))) QUIT(-1);
		tlen+=len+16;

		Message("literal_tbl: %d %d %d %d %d",
			hbit_tbl[0],
			hbit_tbl[1],
			hbit_tbl[2],
			hbit_tbl[3],
			hbit_tbl[4]);

		if (-1==(len=HuffmanMakeEncodingTable(length_mem ,lengths -length_mem ,hencode_tbl+ 256,hbit_tbl+ 256,256))) QUIT(-1);
		tlen+=len;

		Message("lengths: %ld->%ld",lengths-length_mem,len);

		if (-1==(len=HuffmanMakeEncodingTable(hbit_tbl+256,256                 ,hencode_tbl+1296,hbit_tbl+1296, 16))) QUIT(-1);
		tlen+=len+16;
		if (-1==(len=HuffmanMakeEncodingTable(bits_mem   ,bits    -bits_mem   ,hencode_tbl+ 512,hbit_tbl+ 512,256))) QUIT(-1);
		tlen+=len;

		Message("bits: %ld->%ld",bits-bits_mem,len);

		if (-1==(len=HuffmanMakeEncodingTable(hbit_tbl+512,256                 ,hencode_tbl+1312,hbit_tbl+1312, 16))) QUIT(-1);
		tlen+=len+16;
		if (-1==(len=HuffmanMakeEncodingTable(dist_lo_mem,distlo  -dist_lo_mem,hencode_tbl+ 768,hbit_tbl+ 768,256))) QUIT(-1);
		tlen+=len;

		Message("dist-lo: %ld->%ld",distlo-dist_lo_mem,len);

		if (-1==(len=HuffmanMakeEncodingTable(hbit_tbl+768,256                 ,hencode_tbl+1328,hbit_tbl+1328, 16))) QUIT(-1);
		tlen+=len+16;
		if (-1==(len=HuffmanMakeEncodingTable(dist_hi_mem,disthi  -dist_hi_mem,hencode_tbl+1024,hbit_tbl+1024,256))) QUIT(-1);
		tlen+=len;

		Message("dist-hi: %ld->%ld",disthi-dist_hi_mem,len);

		if (-1==(len=HuffmanMakeEncodingTable(hbit_tbl+1024,256                 ,hencode_tbl+1344,hbit_tbl+1344, 16))) QUIT(-1);
		tlen+=len+16;

		Message("Total size: %ld bytes",tlen);

		// write out the file

		llzh.magic = LLZH_MAGIC;
		llzh.header_size = sizeof(struct LossyLZHeader);
		llzh.width = picbm->width;
		llzh.height = picbm->height;
		llzh.literals = literals-literal_mem;
		llzh.runs = lengths-length_mem;

		fwrite(&llzh,sizeof llzh,1,fo);
#if 1
		llzh.literal_pos = ftell(fo);

		HuffmanWriteNibbleTable(fo,hbit_tbl+1280,16);
		HuffmanWriteBlock(fo,hbit_tbl,256,hencode_tbl+1280   ,hbit_tbl+1280   );
		HuffmanWriteBlock(fo,literal_mem,literals-literal_mem,hencode_tbl     ,hbit_tbl      );

		llzh.length_pos = ftell(fo);

		HuffmanWriteNibbleTable(fo,hbit_tbl+1280+16,16);
		HuffmanWriteBlock(fo,hbit_tbl+256,256,hencode_tbl+1280+16,hbit_tbl+1280+16);
		HuffmanWriteBlock(fo,length_mem ,lengths-length_mem  ,hencode_tbl+ 256,hbit_tbl+ 256);

		llzh.bits_pos = ftell(fo);

		HuffmanWriteNibbleTable(fo,hbit_tbl+1280+32,16);
		HuffmanWriteBlock(fo,hbit_tbl+512,256,hencode_tbl+1280+32,hbit_tbl+1280+32);
		HuffmanWriteBlock(fo,bits_mem   ,bits-bits_mem       ,hencode_tbl+ 512,hbit_tbl+ 512);

		llzh.distlo_pos = ftell(fo);

		HuffmanWriteNibbleTable(fo,hbit_tbl+1280+48,16);
Message("distance-lo (1) %08lx",ftell(fo));
		HuffmanWriteBlock(fo,hbit_tbl+768,256,hencode_tbl+1280+48,hbit_tbl+1280+48);
Message("distance-lo (2) %08lx",ftell(fo));
		HuffmanWriteBlock(fo,dist_lo_mem,distlo-dist_lo_mem  ,hencode_tbl+ 768,hbit_tbl+ 768);

		llzh.disthi_pos = ftell(fo);

		HuffmanWriteNibbleTable(fo,hbit_tbl+1280+64,16);
		HuffmanWriteBlock(fo,hbit_tbl+1024,256,hencode_tbl+1280+64,hbit_tbl+1280+64);
		HuffmanWriteBlock(fo,dist_hi_mem,disthi-dist_hi_mem  ,hencode_tbl+1024,hbit_tbl+1024);

	Message("%d %d %d %d %d",*literal_mem,*length_mem,*bits_mem,*dist_lo_mem,*dist_hi_mem);
		fseek(fo,0,SEEK_SET);
		fwrite(&llzh,sizeof llzh,1,fo);
#else
		fwrite(literal_mem,literals-literal_mem,1,fo);
		fwrite(length_mem,lengths-length_mem,1,fo);
		fwrite(bits_mem,bits-bits_mem,1,fo);
		fwrite(dist_lo_mem,distlo-dist_lo_mem,1,fo);
		fwrite(dist_hi_mem,disthi-dist_hi_mem,1,fo);
#endif

	err = tlen;
	}

cleanup0:
	free(literal_mem);
	free(length_mem);
	free(dist_lo_mem);
	free(dist_hi_mem);
	free(bits_mem);
	free(hencode_tbl);
	free(hbit_tbl);
	free(square_tbl);
	return err;
}

#undef QUIT

////////////////////////////////////////////////////////////////////////////

void	LoadLossyLZFile(FILE *fi) {
	struct LossyLZHeader llzh;
	uchar	*literal_mem	=NULL;
	uchar	*length_mem	=NULL;
	uchar	*dist_lo_mem	=NULL;
	uchar	*dist_hi_mem	=NULL;
	uchar	*bits_mem	=NULL;
	uchar	*htable		=NULL;
	uchar	bitbittable[16];
	uchar	bytebittable[256];

	fread(&llzh,sizeof llzh,1,fi);

	Message("bits-size: %ld",(llzh.runs+llzh.literals+7)/8);

	if (!(literal_mem = malloc(llzh.literals))) goto cleanup0;
	if (!(length_mem = malloc(llzh.runs))) goto cleanup0;
	if (!(dist_lo_mem = malloc(llzh.runs))) goto cleanup0;
	if (!(dist_hi_mem = malloc(llzh.runs))) goto cleanup0;
	if (!(bits_mem = malloc((llzh.runs+llzh.literals+6)/8))) goto cleanup0;
	if (!(htable = malloc(4096))) goto cleanup0;

	// this is a major pain
	// load in and decompress each huffman-packed table

	fseek(fi,llzh.literal_pos,SEEK_SET);
	HuffmanReadNibbleTable(fi,bitbittable,16);
	HuffmanReconstructTable(htable,bitbittable,16);
	HuffmanReadBlock(fi,bytebittable,256,htable,bitbittable);

	Message("%d %d %d %d %d",
		bytebittable[0],
		bytebittable[1],
		bytebittable[2],
		bytebittable[3],
		bytebittable[4]);

	HuffmanReconstructTable(htable,bytebittable,256);
	HuffmanReadBlock(fi,literal_mem,llzh.literals,htable,bytebittable);

	Message("%d %d %d %d %d",
		literal_mem[0],
		literal_mem[1],
		literal_mem[2],
		literal_mem[3],
		literal_mem[4]);

	fseek(fi,llzh.length_pos,SEEK_SET);
	HuffmanReadNibbleTable(fi,bitbittable,16);
	HuffmanReconstructTable(htable,bitbittable,16);
	HuffmanReadBlock(fi,bytebittable,256,htable,bitbittable);
	HuffmanReconstructTable(htable,bytebittable,256);
	HuffmanReadBlock(fi,length_mem,llzh.runs,htable,bytebittable);

	fseek(fi,llzh.bits_pos,SEEK_SET);
	HuffmanReadNibbleTable(fi,bitbittable,16);
	HuffmanReconstructTable(htable,bitbittable,16);
	HuffmanReadBlock(fi,bytebittable,256,htable,bitbittable);
	HuffmanReconstructTable(htable,bytebittable,256);
	HuffmanReadBlock(fi,bits_mem,(llzh.literals+llzh.runs+6)/8,htable,bytebittable);

	fseek(fi,llzh.distlo_pos,SEEK_SET);
	HuffmanReadNibbleTable(fi,bitbittable,16);
	HuffmanReconstructTable(htable,bitbittable,16);
Message("distance-lo (1) %08lx",ftell(fi));
	HuffmanReadBlock(fi,bytebittable,256,htable,bitbittable);
	HuffmanReconstructTable(htable,bytebittable,256);
Message("distance-lo (2) %08lx",ftell(fi));
	HuffmanReadBlock(fi,dist_lo_mem,llzh.runs,htable,bytebittable);

	fseek(fi,llzh.disthi_pos,SEEK_SET);
	HuffmanReadNibbleTable(fi,bitbittable,16);
	HuffmanReconstructTable(htable,bitbittable,16);
	HuffmanReadBlock(fi,bytebittable,256,htable,bitbittable);
	HuffmanReconstructTable(htable,bytebittable,256);
	HuffmanReadBlock(fi,dist_hi_mem,llzh.runs,htable,bytebittable);
#if 0
	if (1!=fread(literal_mem,llzh.literals,1,fi)) goto cleanup0;
	if (1!=fread(length_mem,llzh.runs,1,fi)) goto cleanup0;
	if (1!=fread(bits_mem,(llzh.runs+llzh.literals+6)/8,1,fi)) goto cleanup0;
	if (1!=fread(dist_lo_mem,llzh.runs,1,fi)) goto cleanup0;
	if (1!=fread(dist_hi_mem,llzh.runs,1,fi)) goto cleanup0;
#endif
	Message("%d %d %d %d %d",*literal_mem,*length_mem,*bits_mem,*dist_lo_mem,*dist_hi_mem);

	if (llzh.width!=picbm->width || llzh.height!=picbm->height) {
		resize_picture(llzh.width,llzh.height,0);
	};

	Message("Begin decompression!");

	{
		uchar	*literals = literal_mem;
		uchar	*lengths = length_mem;
		uchar	*distlo = dist_lo_mem;
		uchar	*disthi = dist_hi_mem;
		uchar	*bits = bits_mem;
		uchar	*out_ptr = picbm->vm;
		uchar	*limit = picbm->vm + picbm->size;
		long	bitcntr=0;

		*out_ptr++ = *literals++;

		do {
			void fwdmemcpy(uchar *,uchar *,ulong);
			#pragma aux fwdmemcpy="rep movsb" parm [edi] [esi] [ecx] modify exact [edi esi ecx];

			if (bitcntr>=0)
				bitcntr = 0xff000000L + *bits++;

			if (bitcntr&0x80) {
				fwdmemcpy(out_ptr,out_ptr+(ulong)*distlo++ + ((ulong)*disthi++<<8)-4096,*lengths + 3);
				out_ptr += *lengths++ + 3;
				if (!llzh.runs--) {
					Message("Too many runs!");
					return;
				}
			} else {
				*out_ptr++ = *literals++;
				if (!llzh.literals--) {
					Message("Too many literals!");
					return;
				}
			}
			bitcntr<<=1;
		} while(out_ptr<limit);
	}

cleanup0:
	free(literal_mem);
	free(length_mem);
	free(dist_lo_mem);
	free(dist_hi_mem);
	free(bits_mem);
	free(htable);
}

////////////////////////////////////////////////////////////////////////////

typedef	short	HuffInt;

static int HuffmanHistoPassCompare(const HuffInt *i1,const HuffInt *i2);
static void HuffmanBuildTree(ushort *e_table,uchar *bitlen_table,struct HuffmanData *hd_ptr,int maxc);
static void HistoAssignBitlen(int inode,int blen);

static struct HuffmanData {
	ulong	histo[512];
	HuffInt	left[512];
	HuffInt	right[512];
	HuffInt	tmatrix[512];
	HuffInt	bitlcnt[MAX_LEN+1];
	char	bitlen[256];
	char	tdepth[512];

	int	histo_overflow;
} *hdp;

static int HuffmanHistoPassCompare(const HuffInt *i1,const HuffInt *i2) {
	HuffInt n1,n2;

	n1 = *i1;
	n2 = *i2;

	if (hdp->histo[n1]<hdp->histo[n2]) return 1;
	else if (hdp->histo[n1]>hdp->histo[n2]) return -1;
	else if (hdp->tdepth[n1]<hdp->tdepth[n2]) return 1;
	else return -1;
}

ulong	HuffmanMakeEncodingTable(uchar *mem,ulong memsiz,ushort *encode_table,uchar *bitlen_table,int max_code) {
	struct HuffmanData *hd;
	int	i;
	ulong	total_bitlen=0;

	// try to allocate huffmandata struct

	if (!(hd = malloc(sizeof(struct HuffmanData)))) return -1L;

	memset(hd->histo,0,sizeof hd->histo);
	byte_histogram(mem,memsiz,hd->histo);

	HuffmanBuildTree(encode_table,bitlen_table,hd,max_code);

	total_bitlen = 0;
	for(i=0;i<max_code;i++) {
		total_bitlen+=hd->histo[i]*bitlen_table[i];
	}

	free(hd);

	return (total_bitlen+7)>>3;
}

static void	HuffmanBuildTree(ushort *e_table,uchar *bitlen_table,struct HuffmanData *hd_ptr,int maxc) {
	int	i,j;
	int	tnodes,nodes,freeptr,n1,n2,total;
	int	bits;
	ushort	code=0;

	// set globals

	hdp = hd_ptr;

	// initialize all tree nodes

	for(i=0;i<maxc;i++) {
		hdp->left[i]=hdp->right[i]=-1;
		hdp->bitlen[i]=0;
		hdp->tdepth[i]=0;
		bitlen_table[i]=0;
	}

	// search histogram, find out how many nodes we have
	// add each node to translation matrix

	nodes=0;
	for(i=0;i<maxc;i++)
		if (hdp->histo[i]) {
			hdp->tmatrix[nodes] = i;
			++nodes;
		}

	tnodes=total=nodes;
	freeptr = maxc;	// first new free location

	// sort translation matrix

	qsort(hdp->tmatrix,nodes,sizeof hdp->tmatrix[0],HuffmanHistoPassCompare);

	// group least common nodes until we have only one node

	while(nodes>1) {
		n1 = hdp->tmatrix[nodes-1];
		n2 = hdp->tmatrix[nodes-2];
		hdp->left [freeptr] = n1;
		hdp->right[freeptr] = n2;
		hdp->tdepth[freeptr] = max(hdp->tdepth[n1],hdp->tdepth[n2])+1;

		hdp->histo[freeptr] = hdp->histo[n1]+hdp->histo[n2];

		// figure out where we should insert this node into sorted histogram

		i = nodes-2;	// the bottom of the new list
		while(i && hdp->histo[hdp->tmatrix[i-1]]<hdp->histo[freeptr]) --i;

		// shift rest of list down, insert new node

		nodes--;		// shorten list by 1
		memmove(hdp->tmatrix+i+1,hdp->tmatrix+i,sizeof(hdp->tmatrix[0])*(total-i));
		hdp->tmatrix[i] = freeptr;
		++freeptr,++total;
	}

	// assign bitlengths

	hdp->histo_overflow = 0;
	memset(hdp->bitlcnt,0,sizeof(HuffInt)*(MAX_LEN+1));
	HistoAssignBitlen(hdp->tmatrix[0],0);

	memcpy(bitlen_table,hdp->bitlen,sizeof(char)*maxc);
	if (hdp->histo_overflow) {
		do {
			bits = MAX_LEN-1;
			while(!hdp->bitlcnt[bits]) --bits;
			hdp->bitlcnt[bits]	--;
			hdp->bitlcnt[bits+1] += 2;
			hdp->bitlcnt[MAX_LEN]--;
			hdp->histo_overflow-=2;
		} while(hdp->histo_overflow>0);
		for(bits = MAX_LEN;bits;bits--) {
			i = hdp->bitlcnt[bits];
			while(i--) {
				do j=hdp->tmatrix[--total]; while(j>=maxc);
				bitlen_table[j] = bits;
			}
		}
	}

	// construct encoding table

	for(j=1;j<=MAX_LEN;j++) {
		for(i=0;i<maxc;i++) {
			if (bitlen_table[i]==j) {
				e_table[i] = code>>(MAX_LEN-j);

				code += (1L<<MAX_LEN)>>j;

			}
		}
	}
}

static void HistoAssignBitlen(int inode,int blen) {
	if (hdp->left[inode]>-1) {
		HistoAssignBitlen(hdp->left[inode],blen+1);
		HistoAssignBitlen(hdp->right[inode],blen+1);
		if (blen>MAX_LEN) {
			++hdp->histo_overflow;
		}
	} else {
		if (blen>MAX_LEN) {
			hdp->bitlen[inode] = blen = MAX_LEN;
			++hdp->histo_overflow;
		} else {
			hdp->bitlen[inode] = blen;
		}
		hdp->bitlcnt[blen]++;
	};
}

void	HuffmanWriteNibbleTable(FILE *fo,uchar *data,ulong len) {
	while(len>1) {
		fputc(data[0]|(data[1]<<4),fo);
		data+=2;
		len-=2;
	}
	if (len) fputc(data[0],fo);
}

void	HuffmanWriteBlock(FILE *fo,uchar *data,ulong len,ushort *encode_table,uchar *bit_table) {
	uchar	c;
	ulong	accum,taccum;
	uchar	accbits = 0;
	uchar	bits;

	if (!len) return;

	do {
		c = *data++;
		bits = bit_table[c];
		accum <<= bits;
		accbits += bits;
		accum |= encode_table[c];
		if (accbits>=16) {
			taccum = accum>>(accbits-16);
			fputc(taccum,fo);
			fputc(taccum>>8,fo);
			accbits-=16;
		}
	} while(--len);
	if (bits<MAX_LEN) {
		accum <<= (MAX_LEN-bits);
		accbits += (MAX_LEN-bits);
	}
	if (accbits>=16) {
		taccum = accum>>(accbits-16);
		fputc(taccum,fo);
		fputc(taccum>>8,fo);
		accbits-=16;
	}
	if (accbits) {
		taccum = accum<<(16-accbits);
		fputc(taccum,fo);
		fputc(taccum>>8,fo);
	}
}

void	HuffmanReconstructTable(uchar *table,uchar *lengths,ulong maxc) {
	int c1,i,j;

	c1=1L<<(MAX_LEN-1);
	for(j=1;j<=MAX_LEN;j++) {
		for(i=0;i<maxc;i++) {
			if (lengths[i]==j) {
				memset(table,i,c1);
				table += c1;
			}
		}
		c1>>=1;
	}
}

boolean	HuffmanReadNibbleTable(FILE *fi,uchar *table,ulong len) {
	int c;

	do {
		if ((c = getc(fi))==EOF) return FALSE;
		table[0] = c&0x0f;
		table[1] = (c&0xf0)>>4;
		table+=2;

		len-=2;
	} while(len>0);

	return TRUE;
}

boolean	HuffmanReadBlock(FILE *fi,uchar *mem,ulong len,uchar *decode_table,uchar *bitlen) {
	uchar	ch;
	ulong	accum=0,taccum;
	int	accbits=0;

	if (!len) return TRUE;

	do {
		if (accbits<MAX_LEN) {
			int c;

			if (EOF==(c = getc(fi))) return FALSE;
			taccum = (ulong)(uchar)c;
			if (EOF==(c = getc(fi))) return FALSE;
			taccum += ((ulong)(uchar)c)<<8;

			accum |= taccum << (16-accbits);

			accbits += 16;
		}

		*mem++ = ch = decode_table[accum>>(32-MAX_LEN)];

		accbits -= bitlen[ch];
		accum <<= bitlen[ch];
	} while(--len);

	return TRUE;
}

